package mccity.heroes.skills.turret;

import com.herocraftonline.heroes.characters.Hero;
import com.herocraftonline.heroes.characters.party.HeroParty;
import com.herocraftonline.heroes.characters.skill.SkillConfigManager;
import com.herocraftonline.heroes.util.Messaging;
import com.herocraftonline.heroes.util.Setting;
import com.palmergames.bukkit.towny.object.Resident;
import com.palmergames.bukkit.towny.object.Town;
import net.minecraft.server.EntityArrow;
import net.minecraft.server.MathHelper;
import net.minecraft.server.WorldServer;
import org.bukkit.*;
import org.bukkit.block.Block;
import org.bukkit.block.BlockState;
import org.bukkit.block.Dispenser;
import org.bukkit.craftbukkit.CraftWorld;
import org.bukkit.entity.*;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class Turret {

    private static final float Y_OFFSET_OVER_BASE_BLOCK = 1.6f;
    private static final float SCAN_RADIUS = 25;
    private static final int SWITCH_PERIOD_TICKS = 80;

    private final Hero owner;
    private final Block baseBlock;
    private final Location loc;
    private final World world;

    private LivingEntity target = null;
    private long lastSwitchTick = 0;

    private long lastAttackTick = 0;
    private int attackPeriodTicks;

    private float damage;
    private int knockback;
    private boolean fireArrow;

    private int ammo;
    private int ammoPerArrow;

    private long endLifeTick = -1;
    private boolean destroyOnOwnerLogout;
    private boolean destroyOnChunkUnload;

    public Turret(Hero owner, Block base) {
        this.owner = owner;
        this.baseBlock = base;
        this.loc = base.getLocation().add(0.5, Y_OFFSET_OVER_BASE_BLOCK, 0.5);
        this.world = loc.getWorld();
    }

    public void configure(SkillTurret skill) {
        attackPeriodTicks = SkillConfigManager.getUseSetting(owner, skill, "attack-period-ticks", 20, true);
        this.damage = SkillConfigManager.getUseSetting(owner, skill, Setting.DAMAGE, 3, true);
        this.knockback = SkillConfigManager.getUseSetting(owner, skill, "knockback", 0, false);
        this.fireArrow = SkillConfigManager.getUseSetting(owner, skill, "fire-arrow", false);
        ammo = skill.getInitialAmmoFor(owner);
        ammoPerArrow = skill.getAmmoPerArrowFor(owner);

        int lifeTime = skill.getTurretLifeTimeFor(owner);
        if (lifeTime != -1) {
            endLifeTick = world.getFullTime() + lifeTime / 50;
        }
        destroyOnOwnerLogout = SkillConfigManager.getUseSetting(owner, skill, "lifetime.destroy-on-owner-logout", true);
        destroyOnChunkUnload = SkillConfigManager.getUseSetting(owner, skill, "lifetime.destroy-on-chunk-unload", true);
    }

    public Hero getOwner() {
        return owner;
    }

    public String getOwnerName() {
        return owner.getPlayer().getName();
    }

    public void onAiTick() {
        long curTick = world.getFullTime();

        if (ammo <= 0) {
            if (!tryLoadAmmo()) {
                world.playEffect(loc, Effect.SMOKE, 4);
                return;
            }
        }

        if (target != null && (target.isDead() || curTick >= lastSwitchTick + SWITCH_PERIOD_TICKS)) {
            target = null;
        }
        if (target == null) {
            findNewTarget();
            lastSwitchTick = curTick;
        }

        if (target != null && ammo > 0 && curTick >= lastAttackTick + attackPeriodTicks) {
            shootTarget();
            lastAttackTick = curTick;
            ammo--;
            if (ammo == 0) {
                if (!tryLoadAmmo()) {
                    Messaging.send(owner.getPlayer(), Messages.outOfAmmo);
                }
            }
        }
    }

    private boolean tryLoadAmmo() {
        BlockState baseState = baseBlock.getState();
        if (baseState instanceof Dispenser) {
            Dispenser dispenser = (Dispenser) baseState;
            if (Utils.tryConsumeItem(dispenser.getInventory(), Material.ARROW)) {
                ammo += ammoPerArrow;
                return true;
            }
        }
        return false;
    }

    private void findNewTarget() {
        EnderPearl testEntity = world.spawn(loc, EnderPearl.class);
        List<Entity> nearEntities = testEntity.getNearbyEntities(SCAN_RADIUS, SCAN_RADIUS, SCAN_RADIUS);

        List<LivingEntity> possiblyTargets = new ArrayList<LivingEntity>();
        Set<String> friendlyPlayerNames = new HashSet<String>();
        for (Entity nearEntity : nearEntities) {
            if (!(nearEntity instanceof LivingEntity)) continue;
            LivingEntity nearLiving = (LivingEntity) nearEntity;

            if (nearLiving.getHealth() <= 0) continue; // dead

            if (nearLiving instanceof Player) {
                Player nearPlayer = (Player) nearLiving;
                if (nearPlayer.getName().equals(owner.getName())) { // owner
                    friendlyPlayerNames.add(nearPlayer.getName());
                    continue;
                }
                if (nearPlayer.getGameMode() == GameMode.CREATIVE) { // creative
                    friendlyPlayerNames.add(nearPlayer.getName());
                    continue;
                }
                HeroParty party = owner.getParty();
                if (party != null && party.isPartyMember(nearPlayer)) { // party
                    friendlyPlayerNames.add(nearPlayer.getName());
                    continue;
                }
                if (isTownyFriendly(nearPlayer)) { // towny friendly
                    friendlyPlayerNames.add(nearPlayer.getName());
                    continue;
                }
            }

            if (!Utils.inLineOfSight(loc, nearLiving.getLocation().add(0, nearLiving.getEyeHeight(), 0))) continue;

            possiblyTargets.add(nearLiving);
        }

        target = selectTarget(possiblyTargets, friendlyPlayerNames);
    }

    private LivingEntity selectTarget(List<LivingEntity> targets, Set<String> friendlyPlayerNames) {
        if (targets.isEmpty()) return null;

        int targetIndex = -1;
        int maxPriority = 0;

        for (int i = 0; i < targets.size(); i++) {
            LivingEntity target = targets.get(i);
            int curPriority = 0;

            // skip friend's pets
            if (target instanceof Tameable) {
                Tameable pet = (Tameable) target;
                if (pet.getOwner() != null && friendlyPlayerNames.contains(pet.getOwner().getName())) {
                    continue;
                }
            }

            // distance modifier
            double distanceMod = loc.distance(target.getLocation());
            distanceMod = Math.min(Math.max(distanceMod, 1), 24);
            curPriority += (25 - (int) distanceMod);

            // player modifier
            if (target instanceof Player) {
                Player targetPlayer = (Player) target;
                curPriority += 20;
                if (targetPlayer.isSneaking()) curPriority -= 27; // sneaking
                if (owner.isInCombatWith(targetPlayer)) curPriority += 30; // enemy!
            }

            if (curPriority > maxPriority) {
                maxPriority = curPriority;
                targetIndex = i;
            }
        }

        return (targetIndex == -1) ? null : targets.get(targetIndex);
    }

    private boolean isTownyFriendly(Player nearPlayer) {
        if (!SkillTurret.useTowny) return false;

        Resident ownerResident = SkillTurret.townyUniv.getResidentMap().get(owner.getName().toLowerCase());
        Resident nearResident = SkillTurret.townyUniv.getResidentMap().get(nearPlayer.getName().toLowerCase());
        if (ownerResident.hasFriend(nearResident)) return true;

        try {
            if (ownerResident.hasTown()) {
                Town ownerTown = ownerResident.getTown();
                if (nearResident.hasTown()) {
                    Town nearResidentTown = nearResident.getTown();
                    if (ownerTown == nearResidentTown) return true;
                }
            }
        } catch (Exception ex) {
            // never happens
        }

        return false;
    }

    private void shootTarget() {
        WorldServer mcWorld = ((CraftWorld) loc.getWorld()).getHandle();

        EntityArrow mcArrow = new EntityArrow(mcWorld);

        // reset fields setted in the EntityArrow constructor
        mcArrow.width = 0.6F;
        mcArrow.size = net.minecraft.server.Entity.EntitySize.SIZE_2;
        mcArrow.length = 1.8F;

        mcArrow.shooter = null;
        mcArrow.fromPlayer = false;
        Utils.setArrowDamage(mcArrow, damage);
        Utils.setArrowKnockback(mcArrow, knockback);
        if (fireArrow) {
            mcArrow.setOnFire(100);
        }

        // Based on Minecraft aim code (class EntityArrow)
        mcArrow.locY = loc.getY();
        double vecX = target.getLocation().getX() - loc.getX();
        double vecY = target.getLocation().getY() + target.getEyeHeight() - 0.7 - mcArrow.locY;
        double vecZ = target.getLocation().getZ() - loc.getZ();
        double vecXZ = (double) MathHelper.sqrt(vecX * vecX + vecZ * vecZ);
        float yaw = (float) (Math.atan2(vecZ, vecX) * 180.0 / Math.PI) - 90f;
        float pitch = (float) (-(Math.atan2(vecY, vecXZ) * 180.0 / Math.PI));
        double d4 = vecX / vecXZ;
        double d5 = vecZ / vecXZ;
        mcArrow.setPositionRotation(loc.getX() + d4, mcArrow.locY, loc.getZ() + d5, yaw, pitch);
        mcArrow.height = 0f;

        float f4 = (float) vecXZ * 0.2f;
        mcArrow.shoot(vecX, vecY + f4, vecZ, 1.6f, 12f);

        world.playEffect(loc, Effect.BOW_FIRE, 0);

        mcWorld.addEntity(mcArrow);
    }

    public boolean isAlive() {
        boolean alive = endLifeTick == -1 || world.getFullTime() < endLifeTick;
        if (baseBlock.getChunk().isLoaded()) {
            alive = alive && baseBlock.getType() == Material.DISPENSER;
        } else {
            alive = alive && !destroyOnChunkUnload;
        }

        if (owner.getPlayer().isOnline()) {
            alive = alive && owner.canUseSkill("Turret");
        } else {
            alive = alive && !destroyOnOwnerLogout;
        }

        return alive;
    }

    public Block getBaseBlock() {
        return baseBlock;
    }
}
